@@ -16,26 +16,11 @@ BEGIN
     sub OBJECT    () { 512 }
 
     sub HANDLE    () { 16 | 32 }
-
-    if ( $ENV{NO_VALIDATE} )
-    {
-	*validate = sub { 1 };
-	*validate_pos = sub { 1 };
-	*set_options = sub { 1 };
-	*validation_options = sub { 1 };
-    }
-    else
-    {
-	require Params::Validate::Heavy;
-	*validate = \&_validate;
-	*validate_pos = \&_validate_pos;
-	*set_options = \&_validation_options;
-	*validation_options = \&_validation_options;
-    }
 }
 
 require Exporter;
-use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
+
+use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS %OPTIONS $called $options);
 @ISA = qw(Exporter);
 
 my %tags = ( types => [ qw( SCALAR ARRAYREF HASHREF CODEREF GLOB GLOBREF SCALARREF HANDLE UNDEF OBJECT ) ],
@@ -47,7 +32,350 @@ my %tags = ( types => [ qw( SCALAR ARRAYREF HASHREF CODEREF GLOB GLOBREF SCALARR
 @EXPORT_OK = ( @{ $EXPORT_TAGS{all} }, 'set_options' );
 @EXPORT = qw( validate validate_pos );
 
-$VERSION = '0.07';
+$VERSION = '0.08';
+
+# Matt Sergeant came up with this prototype, which slickly takes the
+# first array (which should be the caller's @_), and makes it a
+# reference.  Everything after is the parameters for validation.
+sub validate_pos (\@@)
+{
+    my $p = shift;
+    my @specs = @_;
+
+    if ( $ENV{PERL_NO_VALIDATION} )
+    {
+	foreach my $x (0..$#specs)
+	{
+	    if ( $x > $#{$p} )
+	    {
+		$p->[$x] = $specs[$x]->{default} if ref $specs[$x] && exists $specs[$x]->{default};
+	    }
+	}
+
+	return @$p;
+    }
+
+    # I'm too lazy to pass these around all over the place.
+    local $called = (caller(1))[3];
+    local $options = _get_options( (caller(0))[0] );
+
+    my $min = 0;
+
+    while (1)
+    {
+	last if _is_optional( $specs[$min] );
+	$min++;
+    }
+
+    my $max = scalar @specs;
+
+    my $actual = scalar @$p;
+    unless ($actual >= $min && ( $options->{allow_extra} || $actual <= $max ) )
+    {
+	my $minmax = $options->{allow_extra} ? "at least $min" : ( $min != $max ? "$min - $max" : $max );
+	my $val = $options->{allow_extra} ? $min : $max;
+	$minmax .= $val != 1 ? ' were' : ' was';
+	$options->{on_fail}->( "$actual parameter" . ($actual != 1 ? 's' : '') . " " . ($actual != 1 ? 'were' : 'was' ) . " passed to $called but $minmax expected\n" );
+    }
+
+    foreach ( 0..$#$p )
+    {
+	_validate_one_param( $p->[$_], $specs[$_], "Parameter #" . ($_ + 1) )
+	    if ref $specs[$_];
+    }
+
+    return unless wantarray;
+
+    # Make a copy so we don't alter @_ for the caller.
+    my @p = @$p;
+
+    # Add defaults to existing parameters if called in list context
+    foreach (0..$#specs)
+    {
+	next unless ref $specs[$_];
+	next if $specs[$_]->{optional} || ! exists $specs[$_]->{default};
+
+	$p[$_] = $specs[$_]->{default};
+    }
+
+    return @p;
+}
+
+sub validate (\@$)
+{
+    my $p = shift;
+    my $specs = shift;
+
+    if ( ref $p->[0] && UNIVERSAL::isa( $p->[0], 'HASH' ) )
+    {
+	# Make a copy so we don't alter the hash reference for the
+	# caller.
+	$p = { %{ $p->[0] } };
+    }
+    else
+    {
+	$options->{on_fail}->( "Odd number of parameters in call to $called when named parameters were expected\n" )
+	    if @$p % 2;
+
+	# This is to hashify the list.  Also has the side effect of
+	# copying the values so we can play with it however we want
+	# without actually changing @_.
+	$p = {@$p};
+    }
+
+    if ( $ENV{PERL_NO_VALIDATION} )
+    {
+	while ( my ($key, $spec) = each %$specs )
+	{
+	    if ( ! exists $p->{$key} && ref $spec && exists $spec->{default} )
+	    {
+		$p->{$key} = $spec->{default};
+	    }
+	}
+
+	return %$p;
+    }
+
+    local $called = (caller(1))[3];
+    local $options = _get_options( (caller(0))[0] );
+
+    if ( $options->{ignore_case} || $options->{strip_leading} )
+    {
+	$specs = _normalize_named($specs);
+	$p = _normalize_named($p);
+    }
+
+    unless ( $options->{allow_extra} )
+    {
+	if ( my @unmentioned = grep { ! exists $specs->{$_} } keys %$p )
+	{
+	    $options->{on_fail}->( "The following parameter" . (@unmentioned > 1 ? 's were' : ' was') .
+				   " passed in the call to $called but " .
+				   (@unmentioned > 1 ? 'were' : 'was') .
+				   " not listed in the validation options: @unmentioned\n" );
+	}
+    }
+
+    my @missing;
+    foreach (keys %$specs)
+    {
+	next if exists $p->{$_};
+	push @missing, $_ unless _is_optional($specs->{$_});
+    }
+
+    if (@missing)
+    {
+	my $missing = join ', ', map {"'$_'"} @missing;
+	$options->{on_fail}->( "Mandatory parameter" . (@missing > 1 ? 's': '') . " $missing missing in call to $called\n" );
+    }
+
+    foreach (keys %$specs)
+    {
+	_validate_one_param( $p->{$_}, $specs->{$_}, "The '$_' parameter" )
+	    if ref $specs->{$_} && exists $p->{$_};
+    }
+
+    return unless wantarray;
+
+    # Add defaults to existing parameters if called in list context
+    while (my ($key, $spec) = each %$specs)
+    {
+	next if exists $p->{$key} || ! exists $spec->{default};
+
+	$p->{$key} = $spec->{default};
+    }
+    return %$p;
+}
+
+sub _is_optional
+{
+    my $spec = shift;
+
+    # foo => 1  used to mark mandatory argument with no other validation
+    return ! $spec unless ref $spec;
+
+    return $spec->{optional} if exists $spec->{optional};
+
+    # If it has a default it has to be optional
+    return exists $spec->{default};
+}
+
+sub _normalize_named
+{
+    my $h = shift;
+
+    # we really don't want to mess with the original
+    my %copy = %$h;
+
+    if ( $options->{ignore_case} )
+    {
+	foreach (keys %copy)
+	{
+	    $copy{ lc $_ } = delete $copy{$_};
+	}
+    }
+
+    if ( $options->{strip_leading} )
+    {
+	foreach my $key (keys %copy)
+	{
+	    my $new;
+	    ($new = $key) =~ s/^$options->{strip_leading}//;
+	    $copy{$new} = delete $copy{$key};
+	}
+    }
+
+    return \%copy;
+}
+
+sub _validate_one_param
+{
+    my $value = shift;
+    my $spec = shift;
+    my $id = shift;
+
+    if ( exists $spec->{type} )
+    {
+	my $type = _get_type($value);
+	unless ( $type & $spec->{type} )
+	{
+	    my @is = _typemask_to_strings($type);
+	    my @allowed = _typemask_to_strings($spec->{type});
+	    my $article = $is[0] =~ /^[aeiou]/i ? 'an' : 'a';
+	    $options->{on_fail}->( "$id to $called was $article '@is', which is not one of the allowed types: @allowed\n" );
+	}
+    }
+
+    if ( exists $spec->{isa} )
+    {
+	foreach ( ref $spec->{isa} ? @{ $spec->{isa} } : $spec->{isa} )
+	{
+	    unless ( UNIVERSAL::isa( $value, $_ ) )
+	    {
+		my $is = ref $value ? ref $value : 'plain scalar';
+		my $article1 = $_ =~ /^[aeiou]/i ? 'an' : 'a';
+		my $article2 = $is =~ /^[aeiou]/i ? 'an' : 'a';
+		$options->{on_fail}->( "$id to $called was not $article1 '$_' (it is $article2 $is)\n" );
+	    }
+	}
+    }
+
+    if ( exists $spec->{can} )
+    {
+	foreach ( ref $spec->{can} ? @{ $spec->{can} } : $spec->{can} )
+	{
+	    $options->{on_fail}->( "$id to $called does not have the method: '$_'\n" ) unless UNIVERSAL::can( $value, $_ );
+	}
+    }
+
+    if ($spec->{callbacks})
+    {
+	$options->{on_fail}->( "'callbacks' validation parameter for $called must be a hash reference\n" )
+	    unless UNIVERSAL::isa( $spec->{callbacks}, 'HASH' );
+
+	foreach ( keys %{ $spec->{callbacks} } )
+	{
+	    $options->{on_fail}->( "callback '$_' for $called is not a subroutine reference\n" )
+		unless UNIVERSAL::isa( $spec->{callbacks}{$_}, 'CODE' );
+
+	    $options->{on_fail}->( "$id to $called did not pass the '$_' callback\n" )
+		unless $spec->{callbacks}{$_}->($value);
+	}
+    }
+}
+
+{
+    # if it UNIVERSAL::isa the string on the left then its the type on
+    # the right
+    my %isas = ( 'ARRAY'  => ARRAYREF,
+		 'HASH'   => HASHREF,
+		 'CODE'   => CODEREF,
+		 'GLOB'   => GLOBREF,
+		 'SCALAR' => SCALARREF,
+	       );
+
+    sub _get_type
+    {
+	my $value = shift;
+
+	return UNDEF unless defined $value;
+
+	unless (ref $value)
+	{
+	    # catches things like:  my $fh = do { local *FH; };
+	    return GLOB if UNIVERSAL::isa( \$value, 'GLOB' );
+	    return SCALAR;
+	}
+
+	my $or = 0;
+	if ( ! grep { ref $value eq $_ } keys %isas )
+	{
+	    $or = OBJECT;
+	}
+
+	foreach ( keys %isas )
+	{
+	    return $isas{$_} | $or if UNIVERSAL::isa( $value, $_ );
+	}
+
+	# I really hope this never happens.
+	return UNKNOWN;
+    }
+}
+
+{
+    my %type_to_string = ( SCALAR()    => 'scalar',
+			   ARRAYREF()  => 'arrayref',
+			   HASHREF()   => 'hashref',
+			   CODEREF()   => 'coderef',
+			   GLOB()      => 'glob',
+			   GLOBREF()   => 'globref',
+			   SCALARREF() => 'scalarref',
+			   UNDEF()     => 'undef',
+			   OBJECT()    => 'object',
+			   UNKNOWN()   => 'unknown',
+			 );
+
+    sub _typemask_to_strings
+    {
+	my $mask = shift;
+
+	my @types;
+	foreach ( SCALAR, ARRAYREF, HASHREF, CODEREF, GLOB, GLOBREF, SCALARREF, UNDEF, OBJECT, UNKNOWN )
+	{
+	    push @types, $type_to_string{$_} if $mask & $_;
+	}
+	return @types ? @types : ('unknown');
+    }
+}
+
+{
+    my %defaults = ( ignore_case => 0,
+		     strip_leading => 0,
+		     allow_extra => 0,
+		     on_fail => sub { require Carp;  Carp::confess(shift()) },
+		   );
+
+    sub validation_options
+    {
+	my %opts = @_;
+
+	my $caller = caller;
+
+	foreach ( keys %defaults )
+	{
+	    $opts{$_} = $defaults{$_} unless exists $opts{$_};
+	}
+
+	$OPTIONS{$caller} = \%opts;
+    }
+
+    sub _get_options
+    {
+	my $caller = shift;
+	return exists $OPTIONS{$caller} ? $OPTIONS{$caller} : \%defaults;
+    }
+}
 
 1;
 
@@ -380,6 +708,13 @@ Simple examples of defaults would be:
 
 =head1 USAGE NOTES
 
+=head2 Validation failure
+
+By default, when validation fails C<Params::Validate> calls
+C<Carp::confess>.  This can be overridden by setting the C<on_fail>
+option, which is described in the L<"GLOBAL" OPTIONS|"GLOBAL" OPTIONS>
+section.
+
 =head2 Method calls
 
 When using this module to validate the parameters passed to a method
@@ -434,8 +769,8 @@ module must include a call to C<Params::Validate::validation_options>>.
 =item * ignore_case => $boolean
 
 This is only relevant when dealing with named parameters.  If it is
-true, then the validation code will ignore the case of parameters.
-Defaults to false.
+true, then the validation code will ignore the case of parameter
+names.  Defaults to false.
 
 =item * strip_leading => $characters
 
@@ -467,15 +802,13 @@ The default is to simply use the Carp module's C<confess()> function.
 
 =head1 DISABLING VALIDATION
 
-** This functionality may change in the future **
-
-If the environment variable C<NO_VALIDATION> is set to something true,
-then all calls to the validation functions are turned into no-ops.
-This may be useful if you only want to use this module during
+If the environment variable C<PERL_NO_VALIDATION> is set to something
+true, then all calls to the validation functions are turned into
+no-ops.  This may be useful if you only want to use this module during
 development but don't want the speed hit during production.
 
-I am not terribly happy with the current mechanism for doing this so
-this may change in the future.
+The only error that will be caught will be when an odd number of
+parameters are passed into a function/method that expects a hash.
 
 =head1 LIMITATIONS
 
